/*
  This file is part of CinLife

  CinLife is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  CinLife is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with CinLife.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

#include "CinLife.h"

/* Options to be parsed */
static const char*         options = "x:y:n:h";
static const struct option long_options[] = {
  {"xsize",required_argument,NULL,'x'},
  {"ysize",required_argument,NULL,'y'},
  {"nstep",required_argument,NULL,'n'}  
};

/* Static functions */
static void usage();

/* Debug macros */
#ifndef DRANK
#define DRANK 0
#endif

void initialize_life(local_life* life) {
    
  /* Get rank and size of MPI_COMM_WORLD */

  /* MPI_Dims_create modifies ONLY ZERO ENTRIES */

  /* Create the cartesian topology */

  /* Allocate the local regions and set them to UNDEF */

  /* 
     Definition of MPI Datatypes:
     - mapping of local domains to global domain
     - local domains (excluding halo regions)
     - local columns (for boundary swaps)
     - local rows    (for boundary swaps)
  */

  /* Compute quantities for the base case: no extra column, no extra row */

  /* Mapping from local domains to global domain */
  /* Standard */

  /* Extra row */

  /* Extra column */

  /* Extra row and extra column */

  /* Local domains without halo layers */

  /* Column type on local domains */

  /* Row type on local domains */

}

void set_initial_condition(local_life* life) {
  
  /* Set element in local array to white or black depending on process_coords */

}

void swap_local_boundary(local_life* life) {

  /* Left neighbor */

  /* Right neighbor */

  /* Above neighbor */

  /* Bottom neighbor */

  /* Swap columns with left and right neighbors */

  /* Swap columns with above and bottom neighbors */

  /* Reset boundaries if it is the case */

}

void update_local_domain (local_life* life) {
}

void update_global_domain(local_life* life) {

  if ( life->cartesian_communicator_rank == 0 ) {    

    /* Copy local buffer */

    /* Receive other task buffer */

    /* Compute stride on columns */

    /* Compute stride on rows */

    /* Set startidx at the beginning of the row in the topology, receive data */
      
  } else {
    /* Send local data */
  }
}

void print_local_count(local_life* life) {
  printf("\n");
  for (int ii = 0; ii < life->nrows; ++ii) {
    printf("\n");
    for(int jj = 0; jj < life->ncols; ++jj)
      printf("%3.1d ",life->count[ii*(life->ncols) + jj]);      
  }
  printf("\n\n");
}

void print_to_stdout(const local_life* life) {

  if (life->cartesian_communicator_rank == 0) {
    /* Print buffer for debug purpose */
    for (int ii = 0; ii < life->gsizex*life->gsizey; ++ii) {
      if (ii%life->gsizex == 0)
	printf("\n");
      printf("%3.1d ",life->global_domain[ii]);
    }
    printf("\n\n");
  }
}

void terminate_life(local_life* life) {
  
}

void print_local_region(const local_life* life) {
  printf("\n");
  for (int ii = 0; ii < life->nrows + 2; ++ii) {
    printf("\n");
    for(int jj = 0; jj < life->ncols + 2; ++jj)
      printf("%3.1d ",life->local_domain[ii*(life->ncols + 2) + jj]);      
  }
  printf("\n\n");
}

void read_input_data(int argc,char** argv,local_life* life){

  int opt = 0;
  int idx = 0;
  
  /* Set defaults */
  life->gsizex = DEFAULT_SIZE_X;
  life->gsizey = DEFAULT_SIZE_Y;
  life->nstep  = DEFAULT_NSTEP;
  /* Get options from command line */
  for(;;){

    opt = getopt_long(argc,argv,options,long_options,&idx);
    if (opt == -1) break;

    switch(opt) {
    case 'x':
      life->gsizex = strtol(optarg,NULL,10);
      break;
    case 'y':
      life->gsizey = strtol(optarg,NULL,10);
      break;
    case 'n':
      life->nstep = strtol(optarg,NULL,10);
      break;
    case 'h':
      usage();
      break;
    default:
      break;
    }
  }
}

void print_to_pgm_file(const char* filename, const local_life* life) {

  FILE * outfile;
  if (life->rank == 0) {
    /* Open file */
    outfile = fopen(filename,"wb");
    /* Write header */
    fprintf(outfile,"P5\n");
    /* Write width, height, maximum gray value */
    fprintf(outfile,"%d %d\n%d\n",life->gsizex,life->gsizey,ALIVE);
    /* Write the life board */
    fwrite(life->global_domain,sizeof(char),life->gsizex*life->gsizey,outfile);
    /* Close file */
    fclose(outfile);
  }

}

static void usage() {
    printf("\nmpirun -np [nprocs] ./GameOfLife -x [xsize] -y [ysize] -n [nstep] \n");
    printf("\n");
    printf("-x [xsize] : set the size of the board along x-axis (default: %d)\n",DEFAULT_SIZE_X);
    printf("-y [ysize] : set the size of the board along y-axis (default: %d)\n",DEFAULT_SIZE_Y);
    printf("-n [nstep] : set the number of step to be simulated (default: %d)\n",DEFAULT_NSTEP);
    printf("\n");
    printf("Use your favourite viewer to check the output files! (gqview,xv)\n\n");
}
