/*
  This file is part of CinLife

  CinLife is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.
  
  CinLife is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with CinLife.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <time.h>

#include "mpi.h"
/* 
   CinLife header defines:
   - local_life type
   - functions that operate on it
   - cell_states enum
*/
#include "CinLife.h" 

int main(int argc, char** argv) {

  local_life region;
  char       filename_buffer[50];

  /* Initialize MPI environment */
  MPI_Init(&argc, &argv);

  /* Read size of the global domain */
  read_input_data(argc, argv,&region);

  /* Initialize data structures for the Game of Life */
  initialize_life(&region);
  set_initial_condition(&region);
  /* Print initial condition to file */
  update_global_domain(&region);
#ifdef DEBUG
  print_to_stdout(&region);
#endif
  sprintf(filename_buffer,"life%5.5d.pgm",0);
  print_to_pgm_file(filename_buffer,&region);
  /* Iterate over generations */
  for (int stepidx = 1; stepidx < region.nstep + 1; ++stepidx) {
    /* Iteration step */
    swap_local_boundary(&region);
    update_local_domain(&region);
    update_global_domain(&region);
    /* Print step to file */
#ifdef DEBUG
    print_to_stdout(&region);
#endif
    sprintf(filename_buffer,"life%5.5d.pgm",stepidx);
    print_to_pgm_file(filename_buffer,&region);
  }

  /* Clean the environment (free memory, etc.) before termination */
  terminate_life(&region);
  /* Close MPI environment */
  MPI_Finalize();
  return EXIT_SUCCESS;
}

